const nodefetch = require("node-fetch");
const fs = require("fs");
const https = require("https");

const regex = new RegExp("bitchesgirls.com/(.[^/]+)/(.[^/]+)/(.[^/]+)");
const userRegex = new RegExp("bitchesgirls.com/(.[^/]+)");

class albumDownloader {
    constructor(config) {
        this.config = config;

        if (!config.pics && !config.videos) {
            this.config.pics = true;
            this.config.videos = true;
        }

        if (config.range) {
            let temp = config.range.split("-");
            let min = 0,
                max = 0;
            try {
                min = parseInt(temp[0]);
                max = parseInt(temp[1]);
            } catch (e) {
                this.exit("Range error: " + min + "-" + max + " not useable");
            }

            if (min > max || max < 0 || isNaN(min) || isNaN(max))
                this.exit("Range error: " + min + "-" + max + " not useable");

            this.config.range = { min, max, index: min };
        }
    }

    exit(err) {
        console.error(err);
        process.exit(0);
    }

    async run(url, path, channel) {
        console.log("");
        if (!regex.test(url)) {
            this.exit("Url " + url + " cant be used");
        }

        let links = await this.collectLinks(url, channel);

        if (this.config.i || this.config.info) {
            this.log("Album Info");
            this.log(links.name + "(" + links.altName + ")");
            this.log("url:" + url);
            this.log(links.pics.length + " pics");
            this.log(links.videos.length + " videos");
            return;
        }

        if (!path.endsWith("/")) path = path + "/";

        if (this.config.namefolder && (links.pics.length > 0 || links.videos.length > 0)) {
            path = path + links.altName;
            if (!path.endsWith("/")) path = path + "/";
        }

        if (!fs.existsSync(path)) fs.mkdirSync(path, { recursive: true });

        let pathStatus = fs.lstatSync(path);

        if (!fs.existsSync(path) || !pathStatus.isDirectory()) {
            this.exit("Path " + path + " cant be used");
        }

        let tPath = path;

        if (this.config.seperate) {
            if (!fs.existsSync(tPath + "pics") && this.config.pics) fs.mkdirSync(tPath + "pics");
            if (!fs.existsSync(tPath + "videos") && this.config.videos)
                fs.mkdirSync(tPath + "videos");
        }

        let downloadedPics = await this.downloadFiles(
            links.pics,
            tPath + (this.config.seperate ? "pics" : ""),
            channel
        );
        let downloadedVideos = await this.downloadFiles(
            links.videos,
            tPath + (this.config.seperate ? "videos" : ""),
            channel
        );

        this.clearLine();

        if (!this.config.fast) {
            this.log("Downloaded Album");
            this.log(links.name + "(" + links.altName + ")");
            this.log("url:" + url);
            this.log(
                downloadedPics.count +
                    " pics" +
                    (downloadedPics.skipped ? " (" + downloadedPics.skipped + " Skipped)" : "")
            );
            this.log(
                downloadedVideos.count +
                    " videos" +
                    (downloadedVideos.skipped ? " (" + downloadedVideos.skipped + " Skipped)" : "")
            );
        } else {
            this.writeToChannel(
                " Finished " +
                    links.name +
                    " " +
                    downloadedPics.count +
                    " pics" +
                    (downloadedPics.skipped ? " (" + downloadedPics.skipped + " Skipped)" : "") +
                    " " +
                    downloadedVideos.count +
                    " videos" +
                    (downloadedVideos.skipped ? " (" + downloadedVideos.skipped + " Skipped)" : "")
            );
        }
    }

    async runUser(url, path) {
        let data = await this.collectUserLinks(url);

        this.config.namefolder = true;

        for (let index = 0; index < data.links.length; index++) {
            const link = data.links[index];
            if (!this.config.fast) await this.run(link, path, 0);
            else this.run(link, path, index);
        }
    }

    start(url, path) {
        if (regex.test(url)) {
            this.run(url, path, 0);
        } else if (userRegex.test(url)) {
            this.runUser(url, path);
        }
    }

    collectLinks(url, channel) {
        return new Promise(async (resolve) => {
            let tUrl = url;
            if (!tUrl.endsWith("/")) tUrl = tUrl + "/";
            let urlsplit = url.split("/");

            let data = { pics: [], videos: [], name: "", altName: urlsplit[urlsplit.length - 2] };

            let i = 1;

            while (true) {
                if (this.config.log) this.writeToChannel("Fetching Website " + url + i);

                let Wfetch = await this.fetch(url + i);
                let websiteRaw = await Wfetch.text();
                let parsed = await this.parse(websiteRaw);

                if (this.config.log)
                    this.writeToChannel("Validating: " + (parsed ? "page " + i : "End of Album"));

                if (!parsed) {
                    resolve(data);
                    break;
                }

                if (this.config.pics) data.pics = data.pics.concat(parsed.pics);
                if (this.config.videos) data.videos = data.videos.concat(parsed.videos);
                data.name = parsed.name;
                this.writeToChannel(
                    "Collecting links (" +
                        data.pics.length +
                        " pics " +
                        data.videos.length +
                        " videos)",
                    true
                );

                i++;
            }
        });
    }

    async downloadFiles(links, tPath, channel) {
        return new Promise(async (resolve) => {
            let i = { count: 0, error: 0 };
            for (let index = 0; index < links.length; index++) {
                const link = links[index];

                let temp = link.split("/");
                let name = temp[temp.length - 1];

                
                if (fs.existsSync(tPath + "/" + name) && !this.config.override) {
                    if (!i.skipped) i.skipped = 0;
                    i.skipped++;
                    this.writeToChannel(
                        "Skipping " + name + " (" + (index + 1) + "/" + links.length + ")",
                        true
                    );
                } else {
                    this.writeToChannel(
                        "Downloading " + name + " (" + (index + 1) + "/" + links.length + ")",
                        true
                    );

                    await this.downloadFile(
                        link,
                        tPath,
                        name,
                        (contentLength, bufferLength, fin) => {
                            if (!fin)
                                this.writeToChannel(
                                    "Downloading " +
                                        name +
                                        " (" +
                                        (index + 1) +
                                        "/" +
                                        links.length +
                                        ")(" +
                                        parseInt((bufferLength / contentLength) * 100) +
                                        "%)",
                                    true
                                );
                        },
                        0
                    ).catch((err) => {
                        console.log(" error");
                        i.error++;
                    });

                    i.count++;
                }
            }

            resolve(i);
        });
    }

    async downloadFile(url, path, name, progress, tries) {
        return new Promise(async (resolve, reject) => {
            let alreadyDownloaded = 0;

            let fin = false;

            let stream = fs.createWriteStream(path + "/" + name);

            let urlDATA = new URL(url);

            let options = {
                hostname: urlDATA.hostname,
                path: urlDATA.pathname,
                method: "GET",
                headers: {
                    "User-Agent":
                        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:105.0) Gecko/20100101 Firefox/105.0",
                },
            };

            let req = https.request(options, (res) => {
                res.on("error", (e) => {
                    console.log(e);
                });

                res.on("data", (d) => {
                    stream.write(d);
                    alreadyDownloaded += d.length;
                    if (progress) progress(res.headers["content-length"], alreadyDownloaded, fin);
                });
            });

            req.on("error", (error) => {
                console.log(error);
                stream.end();
                if (tries > 3) reject();
                else {
                    console.log(" weewoo something went wrong retrying...");
                    this.downloadFile(url, path, name, progress, tries + 1)
                        .then((e) => resolve())
                        .catch((e) => reject());
                }
            });

            req.on("close", () => {
                stream.end();
                resolve();
            });

            req.end();
        });
    }

    channelBuffer = {};

    writeToChannel(text, rmline) {
        if (rmline) {
            process.stdout.clearLine(1);

            process.stdout.cursorTo(0);
        } else process.stdout.write("\n");
        process.stdout.write(text);
        //    console.log(text);
    }

    clearLine() {
        if (!this.config.silent && !this.config.log) {
            process.stdout.clearLine();
            process.stdout.cursorTo(0);
        }
    }

    writeLine(line) {
        if (!this.config.silent) process.stdout.write(line);
        if (this.config.log) process.stdout.write("\n");
    }

    log(line) {
        if (!this.config.silent) console.log(line);
    }

    async collectUserLinks(url) {
        let Wfetch = await this.fetch(url);
        let websiteRaw = await Wfetch.text();
        let parsed = await this.parseUser(websiteRaw);

        return parsed;
    }

    parseUser(website) {
        let parts = website.split("<!-- post_template -->");

        let data = { links: [] };

        let reggex = new RegExp('/.+/"');

        for (let index = 1; index < parts.length; index++) {
            let link = parts[index].match(reggex)[0];
            let clean = link.replace('"', "");
            data.links.push("https://bitchesgirls.com" + clean);
        }

        return data;
    }

    parse(website) {
        let json;
        try {
            let split1 = website.split('<script type="application/ld+json">')[1];
            let split2 = split1.split("</script>")[0];
            json = JSON.parse(split2);
        } catch (e) {
            return false;
        }

        let data = { pics: [], videos: [], name: json.headline };

        json.image.forEach((element) => {
            if (!element.url) return;
            // let temp = element.url.split("/");
            // temp.splice(temp.length - 3, 1);
            // data.pics.push(temp.join("/"));
            data.pics.push(element.url);
        });

        json.video.forEach((element) => {
            if (!element.url) return;
           // let temp = element.url.split("/");
           // temp.splice(temp.length - 3, 1);
           
            data.videos.push(element.url);
        });

        return data;
    }

    fetch(url) {
        return new Promise((resovle, reject) => {
            this.runFetch(url, 0, resovle, reject);
        });
    }

    runFetch(url, tries, res, rej) {
        try {
            nodefetch(url).then(
                (resolve) => res(resolve),
                (reject) => {
                    if (tries > 2) rej(reject);
                    else this.runFetch(url, tries + 1, res, rej);
                }
            );
        } catch (e) {
            if (tries > 2) return null;
            return this.runFetch(url, tries + 1, res, rej);
        }
    }
}

module.exports = albumDownloader;
